#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <err.h>
#include <unistd.h>
#include <signal.h>
#include <time.h>

#include "iic.h"
#include "ssd1306.h"
#include "scd30.h"
#include "render.h"

#define IIC_PATH "/dev/iic0"
#define DISPLAY_ADR 0x3c
#define SENSOR_ADR 0x61

#define WIDTH 128
#define HEIGHT 64

/**********************************************************************/

volatile bool finished = false;
void terminate(int sig) {
	finished = true;
}

int main(int argc, char *argv[])
{
	IIC display, sensor;
	uint8_t bits[HEIGHT * WIDTH/8];
	render_t C;
	unsigned vermajor, verminor;
	unsigned ready;
	unsigned meter;
	float offset;
	unsigned interval;
	float forced;
	unsigned automatic;
	float ppm = 0.0, temp = 0.0, humi = 0.0;
	time_t now;
	char tbuf[32];
	int rc;
	int setup;

	if (iic_open(&display, IIC_PATH, DISPLAY_ADR))
		err(1,"open %s", IIC_PATH);

	if (iic_open(&sensor, IIC_PATH, SENSOR_ADR))
		err(1,"open %s", IIC_PATH);

	scd30_soft_reset(&sensor);
	if (scd30_read_version(&sensor, &vermajor, &verminor))
		err(1, "Cannot read firmware version");

	printf("SCD030 Firmware %u.%u\n", vermajor, verminor);

	scd30_get_altitude(&sensor, &meter);
	scd30_get_offset(&sensor, &offset);
	scd30_get_interval(&sensor, &interval);
	scd30_get_forced(&sensor, &forced);
	scd30_get_automatic(&sensor, &automatic);

	printf("SCD030 Altitude:  %u\n", meter);
	printf("SCD030 Offset:    %.2f\n", offset);
	printf("SCD030 Interval:  %u\n", interval);
	printf("SCD030 Forced:    %.2f\n", forced);
	printf("SCD030 Automatic: %u\n", automatic);
	fflush(stdout);

	// Munich
	if (meter != 520)
		scd30_set_altitude(&sensor, 520);

	// Self-Heating
	if (offset != 2.5)
		scd30_set_offset(&sensor, 2.5);

	// Sample interval
	if (interval != 10)
		scd30_set_interval(&sensor, 10);

	if (argc > 1)
		scd30_set_forced(&sensor, atof(argv[1]));

	// Automatic Recalibration
	if (automatic != 1)
		scd30_set_automatic(&sensor, 1);

	signal(SIGINT, terminate);
	signal(SIGTERM, terminate);
	signal(SIGHUP, terminate);

	scd30_start_periodic(&sensor, 0); // pressure

	setup = 1;
	while (!finished) {
		time(&now);

		if (setup) {
			ssd1306_setup(&display, WIDTH, HEIGHT, 0);
			context(&C, bits, WIDTH, HEIGHT);
			setup = 0;
		}

		move(&C, 0, 0);
		rectangle(&C, WIDTH, HEIGHT, true);

		box(&C, WIDTH, HEIGHT, false);

		setfont(&C, "10x20");
		move(&C, 2, 2);

		snprintf(tbuf, sizeof(tbuf), "%6.1f ppm", ppm);
		savepos(&C);
		text(&C, tbuf);
		restorepos(&C);
		relmove(&C, 0, 20);

		setfont(&C, "6x13");
		move(&C, 2, 24);
		savepos(&C);
		draw(&C, WIDTH-4, 0, false);
		restorepos(&C);
		relmove(&C, 20, 2);

		snprintf(tbuf, sizeof(tbuf), "%5.1f degC", temp);
		savepos(&C);
		text(&C, tbuf);
		restorepos(&C);
		relmove(&C, 0, 13);

		snprintf(tbuf, sizeof(tbuf), "%5.1f %%rH", humi);
		savepos(&C);
		text(&C, tbuf);
		restorepos(&C);
		relmove(&C, 0, 13);

		setfont(&C, "5x8");
		move(&C, WIDTH - (2 + 17 * 5), HEIGHT - (2 + 8));

		strftime(tbuf, sizeof(tbuf), "%Y%m%d %H:%M:%S", localtime(&now));
		text(&C, tbuf);

		if (ssd1306_render(&display, bits, WIDTH, HEIGHT))
			setup = 1;

		if (scd30_get_data_ready(&sensor, &ready) || !ready) {
			sleep(1);
			continue;
		}

		rc = scd30_read_measurement(&sensor, &ppm, &temp, &humi);
		if (rc)
			rc = scd30_read_measurement(&sensor, &ppm, &temp, &humi);

		if (rc == 0) {
			printf("%jd %.1f %.1f %.1f\n",
				(intmax_t)now, ppm, temp, humi);
			fflush(stdout);
		}
	}

	scd30_stop_periodic(&sensor);

	iic_close(&sensor);
	iic_close(&display);
}
