#include <string.h>

#include "render.h"

#define min(a,b) ((a) < (b) ? (a) : (b))
#define max(a,b) ((a) > (b) ? (a) : (b))
#define abs(a) ((a) >= 0 ? (a) : -(a))
#define sgn(a) ((a) > 0 ? 1 : (a) < 0 ? -1 : 0)

/* font matrix 16x16 */
#define GLYPHS_PER_LINE 16

#define fontrecord(t, n, w, h) \
	{ (t), (uint8_t *)n##_bits, n##_height, n##_width, (h), (w) }
struct fontrec {
	const char *name;
	const char *bits;
	unsigned height;
	unsigned width;
	unsigned glyph_height;
	unsigned glyph_width;
};

#include "4x6.h"
#include "5x8.h"
#include "6x13.h"
#include "9x15.h"
#include "10x20.h"

static struct fontrec fonts[] = {
	fontrecord("4x6", f4x6, 4, 6),
	fontrecord("5x8", f5x8, 5, 8),
	fontrecord("6x13", f6x13, 6, 13),
	fontrecord("9x15", f9x15, 9, 15),
	fontrecord("10x20", f10x20, 10, 20)
};

static int
setpixel(uint8_t *dst, unsigned dst_stride, int dst_y, int dst_x,
	unsigned v, bool invert)
{
	unsigned off;
	uint8_t mask;

	off = dst_y * dst_stride + dst_x / 8;
	mask = v << dst_x % 8;
	if (invert)
		dst[off] |= mask;
	else
		dst[off] &= ~mask;
}

static int
blit(const uint8_t *src, unsigned src_stride, int src_y, int src_x,
	uint8_t *dst, unsigned dst_stride, int dst_y, int dst_x,
	int height, int width, bool invert)
{
	int y, x;

	for (y=0; y<height; ++y) {
		for (x=0; x<width; ++x) {
			int ya, xa;
			unsigned off, v;
			uint8_t mask;

			if (src) {
				ya = src_y + y;
				xa = src_x + x;
				off = ya * src_stride + xa / 8;
				mask = 1U << xa % 8;
				v = !!(src[off] & mask);
			} else
				v = 1;

			ya = dst_y + y;
			xa = dst_x + x;
			off = ya * dst_stride + xa / 8;
			mask = v << xa % 8;
			if (invert)
				dst[off] |= mask;
			else
				dst[off] &= ~mask;
		}
	}
}

static int
line(uint8_t *dst, unsigned dst_stride, int dst_y, int dst_x,
	int dy, int dx, bool invert)
{
	int error2;
	int adx, ady;
	int sdx, sdy;
	int pdx, pdy;
	int ddx, ddy;
	int slow, fast;
	int i;

	adx = abs(dx);
	sdx = sgn(dx);
	ady = abs(dy);
	sdy = sgn(dy);

	if (adx > ady) {
		pdx = sdx;
		pdy = 0;
		ddx = sdx;
		ddy = sdy;
		slow = ady;
		fast = adx;
	} else {
		pdx = 0;
		pdy = sdy;
		ddx = sdx;
		ddy = sdy;
		slow = adx;
		fast = ady;
	}

	error2 = fast;
	setpixel(dst, dst_stride, dst_y, dst_x, 1, invert);

	for (i=0; i<fast; ++i) {
		error2 -= 2 * slow;
		if (error2 < 0) {
			error2 += 2 * fast;
			dst_x += ddx;
			dst_y += ddy;
		} else {
			dst_x += pdx;
			dst_y += pdy;
		}
		setpixel(dst, dst_stride, dst_y, dst_x, 1, invert);
	}
	return 0;
}

/*********************************************************************/

int
context(render_t *ctx, uint8_t *bits, int width, int height)
{
	memset(ctx, 0, sizeof(*ctx));

	if (width < 0 || height < 0)
		return -1;

	ctx->bits = bits;
	ctx->width = width;
	ctx->height = height;
	ctx->x = 0;
	ctx->y = 0;
	ctx->fontnum = 0;
	return 0;
}

int
move(render_t *ctx, int x, int y)
{
	ctx->x = x;
	ctx->y = y;
	return 0;
}

int
relmove(render_t *ctx, int dx, int dy)
{
	ctx->x += dx;
	ctx->y += dy;
	return 0;
}

int
savepos(render_t *ctx)
{
	ctx->save_x = ctx->x;
	ctx->save_y = ctx->y;
}

int
restorepos(render_t *ctx)
{
	ctx->x = ctx->save_x;
	ctx->y = ctx->save_y;
}

int
setfont(render_t *ctx, const char *name)
{
	unsigned i;
	unsigned n = sizeof(fonts) / sizeof(fonts[0]);

	for (i=0; i<n; ++i) {
		if (strcmp(name, fonts[i].name) == 0) {
			ctx->fontnum = i;
			return 0;
		}
	}
	return -1;
}

int
draw(render_t *ctx, int dx, int dy, bool erase)
{
	unsigned stride = (ctx->width + 7) / 8;

	line(ctx->bits, stride, ctx->y, ctx->x, dy, dx, erase);
	relmove(ctx, dx, dy);
	return 0;
}

int
box(render_t *ctx, int w, int h, bool erase)
{

	draw(ctx, w-1, 0, erase);
	draw(ctx, 0, h-1, erase);
	draw(ctx, 1-w, 0, erase);
	draw(ctx, 0, 1-h, erase);
	return 0;
}

int
rectangle(render_t *ctx, int w, int h, bool erase)
{
	unsigned stride = (ctx->width + 7) / 8;
	int x, y;

	x = min(max(ctx->x, 0), ctx->width);
	y = min(max(ctx->y, 0), ctx->height);

	w = min(w, ctx->width - x);
	h = min(h, ctx->height - y);

	if (w < 0 || h < 0)
		return -1;

	blit(NULL, 0, 0, 0, ctx->bits, stride, y, x, h, w, erase);
	return 0;
}

int
text(render_t *ctx, const uint8_t *str)
{
	int x = ctx->x;
	int y = ctx->y;
	unsigned fontnum = ctx->fontnum;

	const uint8_t *font_bits = fonts[fontnum].bits;
	unsigned glyph_width = fonts[fontnum].glyph_width;
	unsigned glyph_height = fonts[fontnum].glyph_height;
	unsigned font_stride = (fonts[fontnum].width + 7) / 8;
	unsigned font_chars_per_line = fonts[fontnum].width / fonts[fontnum].glyph_width;
	unsigned stride = (ctx->width + 7) / 8;
	int srcy, srcx, dsty, dstx;
	int count = 0;

	dsty = ctx->y;
	dstx = ctx->x;
	while (*str) {

		if (dsty < 0 || dsty + glyph_height > ctx->height)
			break;

		if (dstx < 0 || dstx + glyph_width > ctx->width) {
			dstx = x;
			dsty += glyph_height;
			continue;
		}

		srcy = (*str / font_chars_per_line) * glyph_height;
		srcx = (*str % font_chars_per_line) * glyph_width;

		blit(font_bits, font_stride, srcy, srcx,
			ctx->bits, stride, dsty, dstx,
			glyph_height, glyph_width,
			false);
			
		dstx += glyph_width;

		++str;
		++count;
	}

	ctx->y = dsty;
	ctx->x = dstx;

	return count;
}
