#include <stddef.h>
#include <stdbool.h>
#include <inttypes.h>
#include <errno.h>
#include <unistd.h>

#include "iic.h"
#include "scd30.h"

#define SCD30_CMD_START_PERIODIC_MEASUREMENT 0x0010
#define SCD30_CMD_STOP_PERIODIC_MEASUREMENT 0x0104
#define SCD30_CMD_READ_MEASUREMENT 0x0300
#define SCD30_CMD_SET_MEASUREMENT_INTERVAL 0x4600
#define SCD30_CMD_GET_DATA_READY 0x0202
#define SCD30_CMD_SET_TEMPERATURE_OFFSET 0x5403 
#define SCD30_CMD_SET_ALTITUDE 0x5102
#define SCD30_CMD_SET_FORCED_RECALIBRATION 0x5204
#define SCD30_CMD_AUTO_SELF_CALIBRATION 0x5306
#define SCD30_CMD_READ_FIRMWARE_VERSION 0xD100 
#define SCD30_CMD_READ_SERIAL 0xD033 
#define SCD30_CMD_SOFT_RESET 0xD304 
#define SCD30_SERIAL_NUM_WORDS 16
#define SCD30_WRITE_DELAY_US 20000

#define CRC8_POLYNOMIAL 0x31 
#define CRC8_INIT 0xFF

static
void delay(unsigned us)
{
	/* round up to 10ms */
	usleep(us + 9999);
}

static
uint8_t crc(uint8_t *data, size_t count)
{
	uint16_t current_byte;
	uint8_t crc = CRC8_INIT;
	uint8_t crc_bit;

	/* calculates 8-Bit checksum with given polynomial */
	for (current_byte = 0; current_byte < count; ++current_byte) {
	crc ^= (data[current_byte]);
	for (crc_bit = 8; crc_bit > 0; --crc_bit) {
		if (crc & 0x80)
			crc = (crc << 1) ^ CRC8_POLYNOMIAL;
		else
			crc = (crc << 1);
		}
	}
	return crc;
}

static int
unpack(uint8_t *buf, size_t buflen, uint8_t *out, size_t outlen)
{
	size_t i, k;

	for (i=2, k=0; i<buflen; i+=3, k+=2) {
		if (crc(&buf[i-2], 3))
			return -1;
		if (k+1 < outlen) {
			out[k] = buf[i-2];
			out[k+1] = buf[i-1];
		}
	}
	return k;
}

static float
mkfloat(uint8_t buf[4])
{
	union {
		float f;
		uint32_t i;
	} x;

	x.i = buf[0] << 24
	    | buf[1] << 16
	    | buf[2] << 8
	    | buf[3] << 0;

	return x.f;
}

static int
cmd(IIC *hdl, uint16_t code)
{
	uint8_t cmd[2];
	int rc;

	cmd[0] = code >> 8;
	cmd[1] = code & 0xff;

	rc = iic_write(hdl, cmd, sizeof(cmd), NULL, 0);
	if (rc)
		return rc;

	delay(SCD30_WRITE_DELAY_US);

	return 0;
}

static int
cmd_val(IIC *hdl, uint16_t code, uint16_t val)
{
	uint8_t cmd[2];
	uint8_t args[3];
	int rc;

	cmd[0] = code >> 8;
	cmd[1] = code & 0xff;

	args[0] = val >> 8;
	args[1] = val & 0xff;
	args[2] = crc(args, 2);

	rc = iic_write(hdl, cmd, sizeof(cmd), args, sizeof(args));
	if (rc)
		return rc;

	delay(SCD30_WRITE_DELAY_US);

	return 0;
}

static int
cmd_get(IIC *hdl, uint16_t code, uint16_t *valp)
{
	uint8_t cmd[2];
	uint8_t res[3];
	uint8_t data[2];
	int rc;

	cmd[0] = code >> 8;
	cmd[1] = code & 0xff;

	rc = iic_write(hdl, cmd, sizeof(cmd), NULL, 0);
	if (rc)
		return rc;

	delay(20000);

	rc = iic_read(hdl, NULL, 0, res, sizeof(res));
	if (rc)
		return rc;

	rc = unpack(res, sizeof(res), data, sizeof(data));
	if (rc < 0) {
		errno = EINVAL;
		return -1;
	}

	*valp = data[0] << 8 | data[1];
	return 0;
}

/***************************************************************/

int
scd30_read_serial(IIC *hdl, char *serial, size_t n)
{
	uint8_t res[3 * SCD30_SERIAL_NUM_WORDS];
	int rc;

	rc = cmd(hdl, SCD30_CMD_READ_SERIAL);
	if (rc)
		return rc;

	delay(20000);

	rc = iic_read(hdl, NULL, 0, res, sizeof(res));
	if (rc)
		return rc;

	rc = unpack(res, sizeof(res), serial, n-1);
	if (rc < 0) {
		errno = EINVAL;
		return -1;
	}
	serial[n-1] = '\0';

	return 0;
}

int
scd30_read_version(IIC *hdl, unsigned *vmaj, unsigned *vmin)
{
	uint16_t data;
	int rc;

	rc = cmd_get(hdl, SCD30_CMD_READ_FIRMWARE_VERSION, &data);
	if (rc)
		return rc;

	*vmaj = data >> 8;
	*vmin = data & 0xff;
	return 0;
}

int
scd30_soft_reset(IIC *hdl)
{
	int rc;

	rc = cmd(hdl, SCD30_CMD_SOFT_RESET);
	if (rc == 0)
		delay(500000);

	return rc;
}

int
scd30_set_altitude(IIC *hdl, unsigned meter)
{
	return cmd_val(hdl, SCD30_CMD_SET_ALTITUDE, meter);
}

int
scd30_get_altitude(IIC *hdl, unsigned *meterp)
{
	uint16_t meter;
	int rc;

	rc = cmd_get(hdl, SCD30_CMD_SET_ALTITUDE, &meter);
	if (rc)
		return rc;

	*meterp = meter;
	return 0;
}

int
scd30_set_offset(IIC *hdl, float degrees)
{
	uint16_t off;

	if (degrees < 0.0 || degrees > 655.35) {
		errno = ERANGE;
		return -1;
	}

	off = (uint16_t)(degrees * 100.0);
	return cmd_val(hdl, SCD30_CMD_SET_TEMPERATURE_OFFSET, off);
}

int
scd30_get_offset(IIC *hdl, float *degreesp)
{
	uint16_t data;
	int rc;

	rc = cmd_get(hdl, SCD30_CMD_SET_TEMPERATURE_OFFSET, &data);
	if (rc)
		return rc;

	*degreesp = data / 100.0;
	return 0;
}

int
scd30_set_automatic(IIC *hdl, unsigned enable)
{
	return cmd_val(hdl, SCD30_CMD_AUTO_SELF_CALIBRATION, !!enable);
}

int
scd30_get_automatic(IIC *hdl, unsigned *enablep)
{
	uint16_t data;
	int rc;

	rc = cmd_get(hdl, SCD30_CMD_AUTO_SELF_CALIBRATION, &data);
	if (rc)
		return rc;

	*enablep = data;
	return 0;
}

int
scd30_set_forced(IIC *hdl, float ppm)
{
	uint16_t calib;

	if (ppm < 0.0 || ppm > 5000.0) {
		errno = ERANGE;
		return -1;
	}

	calib = (uint16_t)ppm;
	return cmd_val(hdl, SCD30_CMD_SET_FORCED_RECALIBRATION, calib);
}

int
scd30_get_forced(IIC *hdl, float *ppmp)
{
	uint16_t calib;
	int rc;

	rc = cmd_get(hdl, SCD30_CMD_SET_FORCED_RECALIBRATION, &calib);
	if (rc)
		return rc;

	*ppmp = (float)calib;
	return 0;
}

int
scd30_set_interval(IIC *hdl, unsigned sec)
{
	if (sec < 2 || sec > 1800) {
		errno = ERANGE;
		return -1;
	}

	return cmd_val(hdl, SCD30_CMD_SET_MEASUREMENT_INTERVAL, sec);
}

int
scd30_get_interval(IIC *hdl, unsigned *secp)
{
	uint16_t data;
	int rc;

	rc = cmd_get(hdl, SCD30_CMD_SET_MEASUREMENT_INTERVAL, &data);
	if (rc)
		return rc;

	*secp = data;
	return 0;
}

int
scd30_start_periodic(IIC *hdl, unsigned mbar)
{
	if (mbar != 0 && (mbar < 700 || mbar > 1400)) {
		errno = ERANGE;
		return -1;
	}

	return cmd_val(hdl, SCD30_CMD_START_PERIODIC_MEASUREMENT, mbar);
}

int
scd30_stop_periodic(IIC *hdl)
{
	return cmd(hdl, SCD30_CMD_START_PERIODIC_MEASUREMENT);
}

int
scd30_get_data_ready(IIC *hdl, unsigned *ready)
{
	uint8_t res[3 * 1];
	uint8_t data[2];
	int rc;
	size_t n;

	rc = cmd(hdl, SCD30_CMD_GET_DATA_READY);
	if (rc)
		return rc;

	delay(3000);

	rc = iic_read(hdl, NULL, 0, res, sizeof(res));
	if (rc)
		return rc;

	rc = unpack(res, sizeof(res), data, sizeof(data));
	if (rc < 0 || rc != sizeof(data)) {
		errno = EINVAL;
		return -1;
	}

	*ready = data[0] << 8 | data[1];

	return 0;
}

int
scd30_read_measurement(IIC *hdl, float *ppm, float *temp, float *humi)
{
	uint8_t res[3 * 6];
	uint8_t data[3][4];
	union {
		float f;
		uint32_t i;
	} X;
	int rc;

	rc = cmd(hdl, SCD30_CMD_READ_MEASUREMENT);
	if (rc)
		return rc;

	delay(SCD30_WRITE_DELAY_US);

	rc = iic_read(hdl, NULL, 0, res, sizeof(res));
	if (rc)
		return rc;

	rc = unpack(res, sizeof(res), &data[0][0], sizeof(data));
	if (rc < 0 || rc != sizeof(data)) {
		errno = EINVAL;
		return -1;
	}

	*ppm = mkfloat(&data[0][0]);
	*temp = mkfloat(&data[1][0]);
	*humi = mkfloat(&data[2][0]);

	return 0;
}
