#include <stddef.h>
#include <string.h>
#include <err.h>

#include "iic.h"
#include "ssd1306.h"

#define WIDTH 128
#define HEIGHT 64

enum ssd1306_ctl {
	CMD = 0x00,
	DTA = 0x40,
	MCMD = 0x80,
	MDTA = 0xc0,
};

enum ssd1306_cmd {
	SETCONTRAST = 0x81,
	DISPLAYALLON_RESUME = 0xA4,
	DISPLAYALLON = 0xA5,
	NORMALDISPLAY = 0xA6,
	INVERTDISPLAY = 0xA7,
	DISPLAYOFF = 0xAE,
	DISPLAYON = 0xAF,
	SETDISPLAYOFFSET = 0xD3,
	SETCOMPINS = 0xDA,
	SETVCOMDETECT = 0xDB,
	SETDISPLAYCLOCKDIV = 0xD5,
	SETPRECHARGE = 0xD9,
	SETMULTIPLEX = 0xA8,
	SETLOWCOLUMN = 0x00,
	SETHIGHCOLUMN = 0x10,
	SETSTARTLINE = 0x40,
	SETSTARTPAGE = 0xB0,
	MEMORYMODE = 0x20,
	COLUMNADDR = 0x21,
	PAGEADDR = 0x22,
	COMSCANINC = 0xC0,
	COMSCANDEC = 0xC8,
	SEGREMAP = 0xA0,
	CHARGEPUMP = 0x8D,
	ACTIVATE_SCROLL = 0x2F,
	DEACTIVATE_SCROLL = 0x2E,
	SET_VERTICAL_SCROLL_AREA = 0xA3,
	RIGHT_HORIZONTAL_SCROLL = 0x26,
	LEFT_HORIZONTAL_SCROLL = 0x27,
	VERTICAL_AND_RIGHT_HORIZONTAL_SCROLL = 0x29,
	VERTICAL_AND_LEFT_HORIZONTAL_SCROLL = 0x2A
};

static int
command(IIC *hdl, uint8_t *buf, size_t len)
{
	int rc;

	while (len > 1) {
		switch (*buf) {
		case CMD:
		case DTA:
			rc = iic_write(hdl, NULL, 0, buf, 2);
			if (rc) {
				warn("command");
				return rc;
			}
			buf += 2;
			len -= 2;
			break;
		default:
			err(1, "command: byte %02x", *buf);
			break;
		}
	}
	return 0;
}

static int
image(IIC *hdl, uint8_t *buffer, unsigned rows, unsigned stride)
{
	uint8_t data[] = { DTA };
	size_t len = rows * stride;
	size_t end = len - stride;
	int i, rc = 0;

	for (i=0; i<=end; i += stride) {
		rc = iic_write(hdl, data, sizeof(data), &buffer[i], stride);
		if (rc) {
			warn("image");
			break;
		}
	}
	return rc;
}

int
ssd1306_setup(IIC *hdl, unsigned width, unsigned height, int ext)
{
	uint8_t init[] = {
		CMD, DISPLAYOFF,
		CMD, SETPRECHARGE, CMD, ext ? 0x22 : 0xf0,
		CMD, SETDISPLAYCLOCKDIV, CMD, 0xf0,
		CMD, CHARGEPUMP, CMD, ext ? 0x10 : 0x14,
		CMD, SETVCOMDETECT, CMD, 0x20,
		CMD, SETMULTIPLEX, CMD, height-1,
		CMD, SETDISPLAYOFFSET, CMD, 0,
		CMD, SETSTARTLINE | 0,
		CMD, SEGREMAP | 1,
		CMD, COMSCANDEC,
		CMD, SETCOMPINS, CMD, 0x12,
		CMD, SETCONTRAST, CMD, ext ? 0x9f : 0xcf,
		CMD, DISPLAYALLON_RESUME,
		CMD, NORMALDISPLAY,
		CMD, DEACTIVATE_SCROLL,
		CMD, MEMORYMODE, CMD, 0,
	};
	uint8_t contrast[] = {
		CMD, SETCONTRAST, CMD, 0xff,
	};
	int rc;

	rc = command(hdl, init, sizeof(init));
	if (rc == 0)
		rc = command(hdl, contrast, sizeof(contrast));

	return rc;
}

/* convert xbm bitplane into ssd1306 buffer */
static void
makebuf(uint8_t *bits, unsigned w, unsigned h, uint8_t *buf, size_t len)
{
	unsigned i;

	i = 0;
	for (unsigned y=0; y<h; y+=8) {
		for (unsigned x=0; x<w; ++x) {
			uint8_t byte = 0;
			for (int b=7; b>=0; --b) {
				unsigned row = y + b;
				unsigned off = x / 8;
				unsigned pos = x % 8;
				unsigned bit = (
					bits[row*w/8 + off]
					>> pos
				) & 1;
				byte = byte << 1 | bit;
			}
			if (i >= len)
				break;
			buf[i++] = byte ^ 0xff;
		}
		if (i >= len)
			break;
	}
}

int
ssd1306_render(IIC *hdl, uint8_t *bits, unsigned width, unsigned height)
{
	uint8_t display[] = {
		CMD, DISPLAYON,
		CMD, COLUMNADDR, CMD, 0, CMD, width-1,
		CMD, PAGEADDR, CMD, 0, CMD, height/8-1
	};
	uint8_t buf[HEIGHT * WIDTH/8];

	if (bits)
		makebuf(bits, width, height, buf, sizeof(buf));
	else
		memset(buf, 0, sizeof(buf));

	command(hdl, display, sizeof(display));
	return image(hdl, buf, HEIGHT, WIDTH/8);
}

